# ../addons/eventscripts/gungame51/scripts/custom/gg_dance_off/gg_dance_off.py

'''
        I would just like to extend credit to bonbon for coming up with
    the original disco addon.  Here is the url for the addon:
                http://forums.eventscripts.com/viewtopic.php?f=27&t=23038                  

        I would also like to thank my clan mates and friends for taking the
    time to help me test this thing!  Also, an extended thanks to everyone
    for their compliments, criticisms, translations, and suggestions!

               This addon is released under the "Artistic License"
                  (http://dev.perl.org/licenses/artistic.html)              
'''

# ============================================================================
# >> IMPORTS
# ============================================================================
# Python imports
from __future__ import with_statement
from random import randint
from random import choice
from random import random
from random import shuffle
from os.path import isfile
from path import path
import cPickle

# Eventscripts imports
import es
import repeat
import popuplib
from gamethread import delayed
from effectlib import drawLine
from playerlib import getUseridList
from playerlib import getPlayer
from cmdlib import registerSayCommand
from cmdlib import unregisterSayCommand
from cmdlib import registerServerCommand
from cmdlib import unregisterServerCommand

# Gungame imports
from gungame51.core.addons.shortcuts import AddonInfo
from gungame51.core.players.shortcuts import Player
from gungame51.core.messaging.shortcuts import centermsg
from gungame51.core.weapons.shortcuts import get_total_levels
from gungame51.core.messaging.shortcuts import hudhint
from gungame51.core.messaging.shortcuts import saytext2
from gungame51.core.weapons.shortcuts import get_level_weapon
from gungame51.core.addons import PriorityAddon
    
# ============================================================================
# >> ADDON REGISTRATION/INFORMATION
# ============================================================================
info = AddonInfo()
info.name = 'gg_dance_off'
info.title = 'GG Dance Off' 
info.author = 'Monday (Luke Robinson)' 
info.version = '1.4'
info.conflicts = ['gg_deathmatch']
info.translations = ['gg_dance_off']

# ============================================================================
# >> GLOBAL VARIABLES
# ============================================================================
# Public Version Var
es.ServerVar('gg_dance_off_ver', info.version, info.title).makepublic()

# Settings (Server Vars)
gg_dance_off_time = es.ServerVar('gg_dance_off_time')
gg_dance_off_players = es.ServerVar('gg_dance_off_players')
gg_dance_off_bots = es.ServerVar('gg_dance_off_bots')
gg_dance_off_nade = es.ServerVar('gg_dance_off_nade')
gg_dance_off_knife = es.ServerVar('gg_dance_off_knife')
gg_dance_off_chicken = es.ServerVar('gg_dance_off_chicken')
gg_dance_off_leveldown = es.ServerVar('gg_dance_off_leveldown')
gg_dance_off_levels = es.ServerVar('gg_dance_off_levels')   
gg_dance_off_sound = es.ServerVar('gg_dance_off_sound')
gg_dance_off_chicken_sound = es.ServerVar('gg_dance_off_chicken_sound')

# Other Server Vars
gg_noblock = es.ServerVar('gg_noblock') 
bot_dont_shoot = es.ServerVar('bot_dont_shoot')
dance_index = es.ServerVar('dance_index')
gg_suicide_punish = es.ServerVar('gg_suicide_punish')
eventscripts_gamedir = es.ServerVar('eventscripts_gamedir') 

# Misc Vars
dancers = {}
late_dancers = []
dance_cord = []
bot_dont_shoot_backup = str(bot_dont_shoot)
menu = {'list': [], 'menu': None, 'is_running': False, 'yes': [], 'no': []}
high_scores = {'server_record': [0, 'No records yet!', True]}
db_file_name = es.getAddonPath('gungame51') + ('/scripts/custom/gg_dance_off' +
                                                '/gg_dance_off_scores.db')
priority_adds = []
sound_dir = path('%s/sound' % str(eventscripts_gamedir).replace('\\', '/'))
allow_dance_off = False

# ============================================================================
# >> CLASSES
# ============================================================================
class DancePlayer(object):        
    def __init__(self, userid):
        self.userid = userid
        self.do_view0 = 0
        self.do_view1 = 0
        self.do_location = [0, 0, 0]
        self.do_score = 0.0
        self.name = es.getplayername(self.userid)
        self.do_ask = None
        self.do_vote = 0
        self.do_inVote = False
        self.steamid = Player(userid).steamid
            
    def set_info(self):
        # Get view angles
        self.do_view0 = abs(float(es.getplayerprop(self.userid, 
                            'CCSPlayer.m_angEyeAngles[0]')))
        self.do_view1 = float(es.getplayerprop(self.userid, 
                            'CCSPlayer.m_angEyeAngles[1]')) 
        
        # Get location
        self.do_location = es.getplayerlocation(self.userid)
        
        # Correct view1
        if self.do_view1 < 0:
            self.do_view1 += 360
    
    def update_score(self):
        # New information
        view0 = abs(float(es.getplayerprop(self.userid, 
                            'CCSPlayer.m_angEyeAngles[0]')))
        view1 = float(es.getplayerprop(self.userid, 
                            'CCSPlayer.m_angEyeAngles[1]')) 
        
        location = es.getplayerlocation(self.userid)
        
        # Is player ducked? Give him 10 points
        if getPlayer(self.userid).isDucked():
            self.do_score += 10
        
        # Calculate score
        self.do_score += abs(self.do_view0 - view0) * 0.60
        self.do_score += abs(self.do_view1 - view1) * 0.60
        
        for x in range(3):
            self.do_score += abs(self.do_location[x] - location[x]) * 1.40
                
        # Set new information
        self.do_view0 = view0
        self.do_view1 = view1
        self.do_location = location

# ============================================================================
# >> LOAD & UNLOAD
# ============================================================================
def load():         
    # Does the scores db exist?
    if isfile(db_file_name):
        
        global high_scores
        
        # Open the file, and read it into the dict
        with open(db_file_name, 'r') as DB:
            high_scores = cPickle.load(DB)  

    # Yes/No vote menu
    menu['menu'] = popuplib.easymenu('gg_dance_off', None, dance_menu)
    menu['menu'].menuselect = dance_menu 
            
        # English
    menu['menu'].settitle('Let\'s have a dance off?', lang='en')
    menu['menu'].addoption(True, 'Yes!', 1, 'en')
    menu['menu'].addoption(False, 'No.', 1, 'en')  
    
        # French
    menu['menu'].settitle('Tu veux danser avec moi?', lang='fr')
    menu['menu'].addoption(True, 'Oui!', 1, 'fr')
    menu['menu'].addoption(False, 'Non.', 1, 'fr')  

        # German
    menu['menu'].settitle('Lets dance off haben?', lang='de')
    menu['menu'].addoption(True, 'Ja!', 1, 'de')
    menu['menu'].addoption(False, 'Nicht.', 1, 'de')  

        # Spanish
    menu['menu'].settitle('\xbfDanza de la competencia?', lang='es')
    menu['menu'].addoption(True, '\xa1S\xed!', 1, 'es')
    menu['menu'].addoption(False, 'No.', 1, 'es')  
        
        # Norwegian
    menu['menu'].settitle('Dansekonkurranse?', lang='no')
    menu['menu'].addoption(True, 'Ja!', 1, 'no')
    menu['menu'].addoption(False, 'Ingen.', 1, 'no')  
    
    # Register say command
    registerSayCommand('!danceoff', dance_off_cmd, '!danceoff')
    
    # Unregister bot_dont_shoot as a cheat
    es.server.queuecmd('es_xflags remove cheat bot_dont_shoot')
    
    # Make sounds downloadable
    make_downloadable()
    
    # Allow dance off in 5 seconds
    delayed(5, allow_dance, True)
    
    # Register server command to clear scores
    registerServerCommand('dance_off_clear_scores', 
        clear_scores, 'Command to clear high scores.')
    
    es.dbgmsg(0, 'Loaded: %s' % info.name)

def unload():         
    # Allow gungame to function normally again
    priority(False)
    
    # Did the dance off end early? Then need to stop it!
    if dance_repeat.status() == 2:
        dance_repeat.stop()
        dance_clean(True, True)
        
    # Stop vote if needed
    stop_vote()

    # Delete the menu?
    if menu['menu']:
        menu['menu'].delete()
                                                                                        
    # Save high scores
    save_scores()
    
    # Unregister say command
    unregisterSayCommand('!danceoff')
    
    # Delete the high scores menu if it exists
    if popuplib.exists('gg_dance_off_high'):
        popuplib.delete('gg_dance_off_high')
        
    # Delete the repeater
    repeater = repeat.find('gg_dance_off')
    if repeater != None:
        repeater.delete()    

    # Make sure there is no fog left behind
    es.server.queuecmd('es_xfire %i env_fog_controller ' % es.getuserid() +
                                'turnoff')

    # Change back bot_dont_shoot
    if str(bot_dont_shoot) != bot_dont_shoot_backup:
        es.server.queuecmd('bot_dont_shoot %s' % bot_dont_shoot_backup)
        
    # Register bot_dont_shoot as a cheat again
    es.server.queuecmd('es_xflags add cheat bot_dont_shoot')

    # Unregister server command for clearing scores
    unregisterServerCommand('dance_off_clear_scores')    
    
    es.dbgmsg(0, 'Unloaded: %s' % info.name)
    
# ============================================================================
# >> GAME EVENTS
# ============================================================================
def player_spawn(event_var):
    # If the player didn't join an active team, stop here
    if not event_var['es_userteam'] in ('2', '3'):
        return
    
    userid = int(event_var['userid'])        

    # Is the vote currently running?
    if menu['is_running'] and dance_check():
        menu['menu'].send(userid)
        menu['list'].append(userid)
    
    # Stop the vote
    else:
        stop_vote()
    
    # If no dance is running, we stop here
    if dance_repeat.status() != 2:
        return
    
    # Add player to late spawner list    
    late_dancers.append(userid)
    
    ggPlayer = Player(userid)

    # Send messages to player
    ggPlayer.saytext2(ggPlayer.index, 'LateSpawn', {}, True)
    ggPlayer.saytext2(ggPlayer.index, 'LateSpawn2', {}, True)
    
    # Protect player
    protect(userid)
    
    # Strip player's weapons and switch them to knife
    es.server.queuecmd('es_xsexec %s "use weapon_knife"' % userid)
    delayed(0.25, delayed_strip, userid)
            
def player_death(event_var):
    # Check to see if a dance off needs to be started
    dance_check()
    
    # Remove menu residue player may have left
    userid = int(event_var['userid'])
    delayed(1, menu_remove, userid)

def player_disconnect(event_var):
    # Check to see if a dance off needs to be started
    dance_check()
    
    userid = int(event_var['userid'])
    
    # Remove menu residue player may have left
    delayed(1, menu_remove, userid)

def round_start(event_var):    
    # Allow bots to shoot again?
    if int(bot_dont_shoot) != bot_dont_shoot_backup:
        es.server.queuecmd('bot_dont_shoot %s' % bot_dont_shoot_backup)
        
def round_end(event_var):
    # Allow gungame to function normally again
    priority(False)
    
    # Did the dance off end early? Then need to stop it!
    if dance_repeat.status() == 2:
        dance_repeat.stop()
        dance_clean(True, True)
        
    # Stop vote if needed
    stop_vote()
    
    # Remove prevent level
    for userid in late_dancers:
        ggPlayer = Player(userid)
        if 'gg_dance_off' in ggPlayer.preventlevel:
            ggPlayer.preventlevel.remove('gg_dance_off') 
    
    # Clean out late spawners
    del late_dancers[:]            

def gg_win(event_var):
    # Allow gungame to function normally again
    priority(False)

    # Stop vote if needed
    stop_vote()
    
    # Save high scores
    save_scores()    
                    
    # Did the dance off end early? Then need to stop it!
    if dance_repeat.status() == 2:
        dance_repeat.stop()
        dance_clean(True, True)
        
def es_map_start(event_var):
    # Allow dance off in 15 seconds to prevent premature execution
    # Happens when bots disconnect when players connect
    allow_dance(False)
    delayed(15, allow_dance, True)

# ============================================================================
# >> CUSTOM/HELPER FUNCTIONS
# ============================================================================
def clear_scores(args):
    # Rewrite dict
    global high_scores
    high_scores = {'server_record': [0, 'No records yet!', True]}

    # Save changes
    save_scores()
    
    # Echo the change
    es.server.queuecmd('echo [GG Dance Off] Scores Deleted!!!')    
        
def delayed_strip(userid):
    # Does the player exist?
    if not es.exists('userid', userid):
        return
    
    # Is the player on a team?
    if es.getplayerteam not in (2, 3):
        return
        
    # Is the player dead?
    if getPlayer(userid).isdead:
        return
        
    # Strip them
    Player(userid).strip(True) 

def allow_dance(state):
    # Change the global allow_dance_off
    global allow_dance_off
    allow_dance_off = state

def dance_scores(dancing, return_scores=False):
    # Is there only 1 human?
    if len(dancing) == 1 and not return_scores:
        hudhint('#human,#dead', 'OneScore', 
            {'score': int(dancers[dancing[0]].do_score), 
                'name': es.getplayername(dancing[0])})
        
        hudhint(dancing[0], 'SendScore', 
                {'score': int(dancers[dancing[0]].do_score)})   
        return
        
    # Build score list
    scores = []
    for userid in dancers.keys():
        scores.append([dancers[userid].do_score, 
                                    es.getplayername(userid), userid])
    
    # Sort from highest to lowest
    scores.sort(reverse=True)
    
    # Stop here and return the score list?
    if return_scores:
        return scores
    
    # Send top 2 scores to dead humans
    hudhint('#human,#dead', 'TwoScores', 
            {'score': int(scores[0][0]), 
             'name': scores[0][1],
             'score2': int(scores[1][0]), 
             'name2': scores[1][1]})
             
    # Send scores to dancers   
    for userid in dancing:
        # Is the dancer currently winning?
        if userid == scores[0][2]: 
            hudhint(userid, 'SendScore2', 
                {'score2': int(scores[1][0]), 
                'name': scores[1][1], 
                'score': int(scores[0][0])})
            continue
            
        # Send user his own score and the top score
        hudhint(userid, 'SendScore2', 
                {'score2': int(scores[0][0]), 
                'name': scores[0][1], 
                'score': int(dancers[userid].do_score)})        

def make_downloadable():
    # Is the var valid?
    if str(gg_dance_off_sound) not in ('', '0', None, 'None'):
        # Does the file exist?
        if isfile(sound_dir.joinpath(str(gg_dance_off_sound))):
            # Make downloadable
            es.stringtable('downloadables', ('sound' +
                                            '/%s') % str(gg_dance_off_sound))              

    # Is var valid?
    if str(gg_dance_off_chicken_sound) not in ('', '0', None, 'None'):
        # Does the file exist?
        if isfile(sound_dir.joinpath(str(gg_dance_off_chicken_sound))):
            # Make downloadable
            es.stringtable('downloadables', ('sound' +
                                    '/%s') % str(gg_dance_off_chicken_sound))

def dance_off_cmd(userid, args):
    # Don't allow restart if active
    if dance_repeat.status() == 2:
        return
        
    # Did the player close the menu?
    if userid in menu['list']:
        menu['menu'].send(userid)
    
    # Did the player already vote yes?
    elif userid in menu['yes']:
        menu['list'].append(userid)
        menu['yes'].remove(userid)
        menu['menu'].send(userid)
    
    # Did the player already vote no?
    elif userid in menu['no']:
        menu['list'].append(userid)
        menu['no'].remove(userid)
        menu['menu'].send(userid)
    
    # Send player the high score list    
    else:
        # Build list
        scores = high_scores.values()

        # Sort high to low
        scores.sort(reverse=True)
        
        # Convert to string w/ names (top 50)
        scores = ['%s - %s' % (a[0], a[1]) for a in scores[:49] if len(a) != 3]
        
        # Delete the menu if it allready had existed
        if popuplib.exists('gg_dance_off_high'):
            popuplib.delete('gg_dance_off_high')            
        
        # Is the list empty? Send no record message
        if scores == []:
            scores = ['No records have been set!']
        
        high_score_menu = popuplib.easylist('gg_dance_off_high', scores)        
        
        # Titles in different languages
        high_score_menu.settitle('GG Dance Off Scores:', lang='en')
        high_score_menu.settitle('GG Dance Off Scores:', lang='fr')
        high_score_menu.settitle('GG Dance Off Partituren:', lang='de')
        high_score_menu.settitle('GG Dance Off Resultados:', lang='es')
        high_score_menu.settitle('GG Dance Off Scorene:', lang='no')
        
        # Send the menu
        high_score_menu.send(userid)

def save_scores():
    # Write the high scores dict to file
    with open(db_file_name, 'w') as DB:
        cPickle.dump(high_scores, DB)  

def stop_vote():
    # Reset menu
    del menu['yes'][:]
    del menu['no'][:]
    del menu['list'][:]

    # Is it currently running?
    if not menu['is_running']:
        return
    
    menu['is_running'] = False

    # Unsend the menu
    menu['menu'].unsend(getUseridList('#human'))
    
def dance_menu(userid, choice, popupname=None):
    # Don't allow restart if active
    if dance_repeat.status() == 2:
        return

    # Make sure they can vote
    if userid not in menu['list']:
        return
    
    pPlayer = getPlayer(userid)
        
    # Make sure the player isn't dead or a spec
    if pPlayer.isdead or es.getplayerteam(userid) < 2:        
        # Remove the userid from any/all menu lists
        menu_remove(userid)
        return       

    # No
    if choice != True:
        menu['list'].remove(userid)
        menu['no'].append(userid)
        
        # Player is a chicken?
        if int(gg_dance_off_chicken):
            
            # Play sound to humans?
            if str(gg_dance_off_chicken_sound) not in ('', '0', None, 'None'):
                for player in getUseridList('#human'):
                    es.server.queuecmd('es_xcexec %s ' % player +
                            'play %s' % str(gg_dance_off_chicken_sound))
            
            index = Player(userid).index
            
            # Send message about the chicken
            saytext2('#human', index, 'Chicken', 
                                {'name': es.getplayername(userid)}, True)            
        
        # Send command message
        index = Player(userid).index        
        Player(userid).saytext2(index, 'Command', {}, True)
        
        return        
    
    # Yes
    menu['list'].remove(userid)
    menu['yes'].append(userid)           
    
    # Can we start yet?
    dance_start_check(userid)
    
def dance_start_check(userid):
    # Someone says no ?
    if menu['no']:
        return
        
    # Someone still needs to vote ?
    if menu['list']:
        return

    # Making sure there was at least 1 vote
    if not menu['yes']:
        return
        
    humans = getUseridList('#human, #alive') 
    
    # Make sure there is at least 1 human player    
    if not humans:
        return
        
    # Start dance!
    start_dance_loop(userid)
    menu['is_running'] = False

def menu_remove(userid):
    # Stop here if the dance is allready running
    if dance_repeat.status() == 2:
        return
        
    # Remove userid from yes list and stop here?
    if userid in menu['yes']:
        menu['yes'].remove(userid)
        return
        
    # Remove userid from list?
    if userid in menu['list']:
        menu['list'].remove(userid)
        
    # Remove userid from no list?
    elif userid in menu['no']:
        menu['no'].remove(userid)
    
    # Player was not in any lists? Stop here.    
    else:
        return
        
    # Recheck to see if the vote is go?
    dance_start_check(userid)        

def dance_check():
    # Allow dance offs right now?
    if not allow_dance_off:
        return False

    # Make sure a dance is not currently active
    if dance_repeat.status() == 2: 
        return False
    
    human_list = getUseridList('#alive, #human') 
    
    # Are there any human players alive? If not stop here
    if not len(human_list):
        return False
        
    # If there is at least 1 player in CT
    if not len(getUseridList('#ct, #alive')) >= 1:
        return False
    
    # If there is at least 1 player in T
    if not len(getUseridList('#t, #alive')) >= 1:
        return False

    bot_list = getUseridList('#alive, #bot')

    # Are bots allowed? If not stop here
    if len(bot_list) and not int(gg_dance_off_bots):
        return False
    
    # Has the threshold been met? If not stop here
    if len(bot_list + human_list) > int(gg_dance_off_players):
        return False
        
    # Has the vote allready gone out?
    if len(menu['yes'] + menu['no'] + menu['list']):
        return False
    
    # Is a vote currently in progress?
    if menu['is_running']:
        return True
    
    # Set list
    menu['list'].extend(human_list)    
    
    # Everything checked out, starting vote
    menu['menu'].send(human_list)
    
    # Anwser for all bots
    menu['yes'].extend(bot_list)
    
    # Set the vote to running
    menu['is_running'] = True      

def protect(userid=None):
    # All alive players?
    if not userid:
        userid_list = getUseridList('#alive')
    
    # Single player
    else:
        userid_list = [userid]
    
    # Loop through players
    for userid in userid_list:        
        pPlayer = getPlayer(userid)
        
        # Make sure player is alive and able
        if pPlayer.isdead or pPlayer.teamid < 2:
            return
        
        # Make the player invinsible if they are not allready so
        if not pPlayer.inGodMode(): 
            pPlayer.godmode(1)
        
        # NoBlock the player if needed
        if not int(gg_noblock) and not pPlayer.getNoBlock():
            pPlayer.noblock(1)

def start_dance_loop(userid):   
    # Protect the players (noblock/god mode)
    protect()
    
    # Clear dancer dict
    dancers.clear()
    
    # Get player cords
    x, y, z = es.getplayerlocation(userid)
    
    # Add to get z height above the player's head
    z += 125
    	
    # Create ball
    es.server.queuecmd('es_xprop_dynamic_create %s roller.mdl' % userid)  
    es.server.queuecmd('es_setindexprop server_var(eventscripts_lastgive) ' +
                        '"CBaseEntity.m_vecOrigin" ' +
                        '"%s"' % ','.join(map(str, [x, y, z])))
    
    # Get the index
    es.server.queuecmd('es_xcopy dance_index eventscripts_lastgive')
   
    # Create fog controller
    es.server.queuecmd('es_xgive %s env_fog_controller' % userid)
    es.server.queuecmd('es_xfire %s env_fog_controller addoutput ' % userid +
                        '"fogdir 180 180 90"')
    es.server.queuecmd('es_xfire %s env_fog_controller addoutput ' % userid +
                        '"fogstart 155"')
    es.server.queuecmd('es_xfire %s env_fog_controller addoutput ' % userid +
                        '"fogend 500"')
    es.server.queuecmd('es_xfire %s env_fog_controller addoutput ' % userid +
                        '"fogblend 1"')
    es.server.queuecmd('es_xfire %s env_fog_controller addoutput ' % userid +
                        '"fogcolor2 255 255 0"')
    es.server.queuecmd('es_xfire %s env_fog_controller addoutput ' % userid +
                        '"fogcolor 0 0 255"')
    es.server.queuecmd('es_xfire %s env_fog_controller turnon' % userid)
    
    # Start loop
    delayed(0.50, dance_repeat.start, 
                                (0.10, int((gg_dance_off_time + 4) / 0.10)))
    
    # Set global cord of the ball
    global dance_cord
    dance_cord = [x, y, z] 

    z -= 100
    
    # Start music?
    if str(str(gg_dance_off_sound)) not in ('', '0', None, 'None'):
        for player in getUseridList('#human'):        
                    es.server.queuecmd('es_xcexec %s ' % player +
                            'play %s' % str(gg_dance_off_sound))
        
    # Start teleports
    for player in getUseridList('#alive'):      
        
        # Don't teleport base player
        if userid == player:
            continue
            
        # Don't teleport bots
        if es.isbot(userid):
            continue
            
        # Teleport
        es.delayed(1.50, 'es_xsetpos %s %s %s %s' % (player, x, y, z))        
        
    # Disable shooting for bots
    if str(bot_dont_shoot) != '1':
        es.server.queuecmd('bot_dont_shoot 1')
        global bot_dont_shoot_backup
        bot_dont_shoot_backup = 0
        
    # Prevent everything but dance off from running right now
    priority(True)

def priority(status):
    # Prevent everything but dance off from running right now?
    if status and 'gg_dance_off' not in PriorityAddon():
        PriorityAddon().append('gg_dance_off')
    
        # No block?
        if int(gg_noblock):
            PriorityAddon().append('gg_noblock')    
            priority_adds.append('gg_noblock')    

        # Suicide Punish?
        if int(gg_suicide_punish) and int(gg_dance_off_leveldown):
            PriorityAddon().append('gg_suicide_punish') 
            priority_adds.append('gg_suicide_punish')   
            
    # Allow gungame to run normally?
    elif not status and 'gg_dance_off' in PriorityAddon():
        PriorityAddon().remove('gg_dance_off')

        # Remove any added addons
        for addon in priority_adds:
            PriorityAddon().remove(addon)            

def dance_clean(smoke=True, stopped=False):
    # Remove ball
    if dance_index in es.createentitylist().keys():
        es.server.queuecmd("es_xremove %s" % dance_index)
        
    # Remove smoke    
    if smoke:
        userid = es.getuserid()
        es.server.queuecmd('es_xfire %i env_fog_controller ' % userid +
                                'turnoff')
    	es.server.queuecmd('es_xfire %i env_fog_controller kill' % userid)
        
    # Declare winner
    if stopped:
        dance_finish(True)           

def dance_finish(stopped=False):
    alive = getUseridList('#alive')
    
    # Get score list
    scores = dance_scores([], True)
    
    # Process the scores and record new records
    for score in scores:        
        # Was a late spawner?
        if score[2] in late_dancers:
            scores.remove(score)
            continue
            
        # Is player still alive and on the server?
        if score[2] not in alive:
            scores.remove(score)
            continue
            
        steamid = dancers[score[2]].steamid
        
        # Save high score?
        if steamid in high_scores.keys():
            # Beat the old high score?
            if high_scores[steamid][0] < score[0]:
                new_high_score = True
            
            # Old score stands
            else:
                new_high_score = False
        
        # Never recorded a score.. Lets record one!
        else:
            new_high_score = True
            
        # New high score?
        if new_high_score:
            high_scores[steamid] = [int(score[0]), score[1]]
            ggPlayer = Player(score[2])
            index = ggPlayer.index
            ggPlayer.saytext2(index, 'HighScore', 
                {'score': int(score[0])}, True)                        
    
    # Check for the possibility no valid scores were recorded
    if scores == []:
        return

    # Allow gungame to function normally again
    priority(False)
    
    ggPlayer = Player(scores[0][2])

    levels = int(gg_dance_off_levels)
    stop_loop = False
    total = get_total_levels()
    
    # Loop through levels to see how high the player can level
    for level in range(1, (levels + 1)):
        # Does the next level exist?
        if (ggPlayer.level + level) > total:
            level = (total - ggPlayer.level)
            stop_loop = True         
        
        # Next level knife?
        if get_level_weapon(ggPlayer.level + level - 1) == 'knife':
            # Can level past knife?
            if not int(gg_dance_off_knife):
                level -= 1
                break
                
        # Next level nade?
        if get_level_weapon(ggPlayer.level + level - 1) == 'hegrenade':
            # Can level past nade?
            if not int(gg_dance_off_nade):
                level -= 1
                break
                
        # Stop?
        if stop_loop:
            break
            
    # Make sure level is not negitive
    if level < 0:
        level = 0    
                
    # Level up the player?
    if level:
        ggPlayer.levelup(level, 0, 'gg_dance_off')
        ggPlayer.playsound('levelup')                
    
    # Send can't level message?
    if level != levels:
        ggPlayer.saytext2(ggPlayer.index, 'NoLevel', 
                {'weapon': ggPlayer.weapon}, True)                  
       
    # Send winner msg
    saytext2('#human', ggPlayer.index, 'WinnerMsg', 
        {'name': scores[0][1], 'score': int(scores[0][0])}, True)    
    
    # New server record?
    if int(scores[0][0]) > high_scores['server_record'][0]:
        # Set new server record
        high_scores['server_record'][0] = int(scores[0][0])
        high_scores['server_record'][1] = scores[0][1]
        
        # Send message to server
        saytext2('#human', ggPlayer.index, 'WorldRecord', 
            {'name': scores[0][1]}, True)
            
        # Save high scores since a new server record was achieved
        save_scores()                              

    # Early stop?
    if stopped:
        return 
    
    # Slay the losers
    for userid in alive:
        
        # Dont slay the winner
        if userid == scores[0][2]:
            continue
        
        # Level down?
        if int(gg_dance_off_leveldown) and not int(gg_suicide_punish):
            ggPlayer = Player(userid)
            
            # Can player level down?
            if ggPlayer.level > 1 and userid not in late_dancers:
                ggPlayer.level -= 1
                ggPlayer.playsound('leveldown')
            
        # Protect late spawners if we are leveling down
        elif int(gg_dance_off_leveldown) and int(gg_suicide_punish):
            if userid in late_dancers:
                Player(userid).preventlevel.append('gg_dance_off')                   

        # Slay
        es.server.queuecmd('es_xsexec %s kill' % userid)
        
        # Correct score (add frag back)
        es.server.queuecmd('score add %s 1' % userid)

def dance_loop():
    userid_list = getUseridList('#alive, #human')
    bot_list = getUseridList('#alive, #bot')
    
    # No alive humans? Then we need to stop the dance off now!
    if not userid_list:
        dance_repeat.stop()
        dance_clean()
        
        # Slay the bots
        for bot in bot_list:
            es.server.queuecmd('es_xsexec %s kill' % bot)
        
        return
    
    # Make sure the player has a class assigned
    for userid in userid_list:
        if userid not in dancers.keys():
            dancers[userid] = DancePlayer(userid)
    
    timeleft = dance_repeat['timeleft'] 

    # Grab last minute info, and tell everyone to boogie ?
    if float(timeleft - int(gg_dance_off_time)) == 0.0:
        # Set initial information about the player
        for userid in userid_list:
            doPlayer = dancers[userid]
            doPlayer.set_info()
        
        centermsg('#human', 'GoTime', {})     

    # Count down ?
    elif timeleft >= int(gg_dance_off_time) and timeleft % 1 == 0:
        centermsg('#human', 'StartingIn', {'time': (int(timeleft) - 
                                                    int(gg_dance_off_time))})   

    # effects are good to go
    else:
        # Human loop
        for userid in userid_list:
            # Update player's score
            doPlayer = dancers[userid]
            doPlayer.update_score()           
                
        # Bot loop
        for bot in bot_list:
            es.server.queuecmd('es_xsetang %s ' % bot +
                           '%s %s 0' % (randint(0, 360), randint(0, 360)))            

            # Find out where the bot went off to
            bot_cord = es.getplayerlocation(bot)
            
            # Calculate the traveled distance
            _bot_dist = abs(dance_cord[0] - bot_cord[0])
            _bot_dist += abs(dance_cord[1] - bot_cord[1]) 
            
            # If the bot has wandered off, then teleport him back to the ball
            f, g, h = dance_cord
            h -= 110
            if _bot_dist > 500:
                es.server.queuecmd('es_xsetpos %s ' % bot + 
                                        '%s %s %s' % (f, g, h))                 
       
        # Count Down ?
        if 0 < timeleft <= 3:
            centermsg('#human,#alive', 'TimeLeft', {'time': timeleft})
            
        # Send stop message ?
        elif float(timeleft) == 0.0:
            centermsg('#human,#alive', 'Stop', {})
            delayed(0.75, dance_clean)
            dance_finish()   
          
        # Send Scores?
        if not timeleft % 0.50:
            dance_scores(userid_list)  
    
    # Axis of the ball
    dance_axis = es.getindexprop(dance_index, 'CBaseEntity.m_angRotation')
    
    # Make sure it came back as a string, and not something else
    if type(dance_axis) != str:
        # Allow gungame to function normally again
        priority(False)
        
        # Stop the dance off, something is wrong
        dance_repeat.stop()
        dance_clean(True, True)
        
        return
              
    # Convert str into a list
    dance_axis = list(map(float, dance_axis.split(',')))

    # Get a end point for the laser (player)
    if not randint(0, 5):
        p = list(es.getplayerlocation(choice(userid_list)))        

    # Get a end point for the laser (random)
    else:
        p = [(x + (randint(50, 350) * choice((-1,1)))) for x in dance_cord]

    # Effects loop         
    for i in range(5):
        # Grab set of colors
        rgb = colors()
        
        # Set destination
        p2 = [x + randint(-90, 90) for x in p] 
        
        # Draw new laser
        drawLine(dance_cord, p2, "materials/sprites/laser.vmt", 
          "materials/sprites/halo01.vmt", max(random(), 0.40), randint(5, 30), 
          randint(10, 45), rgb[0], rgb[1], rgb[2], 255, 10, 0, 0, 0, 0)
          
        # Move ball axis
        if i == 1:
            dance_axis[1] += 12.50                    
    
        # Change smoke color every 1.5 seconds
        if not dance_repeat['count'] % 15:
            
            # Change the smoke color for the 2nd controller
            if i == 2:
                rgb.insert(0, 2)
            
            # Change the smoke color for the main controller
            elif i == 0:
                rgb.insert(0, '')
            
            # No change is needed
            else:
                continue            
            
            # Change color of fog
            es.server.queuecmd('es_xfire %i ' % es.getuserid() +
                'env_fog_controller addoutput ' +
                '\"fogcolor%s %i %i %i\"' % tuple(rgb))           

    # Move ball
    es.setindexprop(dance_index, 'CBaseEntity.m_angRotation', 
                '%d,%d,%d' % tuple(dance_axis))   

def colors():
    color_selection = (0, 75, 125, 255)    
    color = [0, 0, 0]
	
	# Pick any color but black
    while sum(color) == 0:
	   color = [choice(color_selection) for x in (1, 2, 3)]
	   
    return color
    
# ============================================================================
# >> REPEATER
# ============================================================================

'''
    This was done for insurance, as sometimes unloads/loads with unlucky
    timing would crash a server. :(  I added the global to the end of the file
    so the function it points to is not loaded at the begining.
'''

# Does the loop already exists?
loop = repeat.find('gg_dance_off')
if loop != None:
    
    # Is the loop running? Then stop it!
    if loop.status() == 2:
        loop.stop()
        
        userid = es.getuserid()
        
        # Kill the fog controller
        es.server.queuecmd('es_xfire %i env_fog_controller ' % userid +
                            'turnoff')
        es.server.queuecmd('es_xfire %i env_fog_controller kill' % userid)

        # Slay all terrorists to end the round
        for userid in getUseridList('#alive, #t'):
            es.server.queuecmd('es_xsexec %s kill' % userid)
            
    # Delete the old loop
    loop.delete()

# Delete vars
del loop
if 'userid' in locals().keys():
    del userid

# Global for the loop
dance_repeat = repeat.create('gg_dance_off', dance_loop)
    
